import argparse
import re
import sys
import traceback
from typing import Any, Iterator, List, Optional, Union


# TODO: for this function, I would like to have a decorator that splits the given code_text into code blocks for each function and then creates a sig for each of them
# @decorators.map_firstp_arg
def python_functions_signatures(
    code_text: str, *, ignore_private_functions: bool = False, keep_function_name: bool = False
) -> List[str]:
    """Return the function signatures (the parameters and their types/default values) for all of the functions in the given code_text."""
    from d8s_strings import string_remove_from_start

    from .ast_data import python_function_names

    signatures = []

    function_names = python_function_names(code_text, ignore_private_functions=ignore_private_functions)

    for name in function_names:
        regex_for_signature = f'(def {name}\((?:.|\s)*?\).*?):'
        sig = re.findall(regex_for_signature, code_text)
        if any(sig):
            new_sig = string_remove_from_start(sig[0], 'def ')
            if not keep_function_name:
                new_sig = string_remove_from_start(new_sig, name)
            signatures.append(new_sig)
        else:
            message = f'Unable to find signature for the {name} function'
            print(message)
            signatures.append(None)

    return signatures


def python_todos(code_text: str, todo_regex: str = 'TODO:.*') -> List[str]:
    """Return all todos in the given code_text that match the given todo_regex."""
    todos = re.findall(todo_regex, code_text)
    return todos


# @decorators.map_first_arg
def python_make_pythonic(name: str) -> str:
    """Make the name pythonic (e.g. 'fooBar' => 'foo_bar', 'foo-bar' => 'foo_bar', 'foo bar' => 'foo_bar', 'Foo Bar' => 'foo_bar')."""
    from d8s_strings import lowercase, snake_case, string_split_on_uppercase

    split_string = '_'.join(
        [
            string.strip()
            for string in string_split_on_uppercase(name, include_uppercase_characters=True, split_acronyms=False)
        ]
    )
    split_string = lowercase(split_string)
    result = snake_case(split_string)
    return result


# @decorators.map_first_arg
def python_namespace_has_argument(namespace: argparse.Namespace, argument_name: str) -> bool:
    """."""
    result = argument_name in namespace
    return result


# @decorators.map_first_arg
def python_traceback_prettify(traceback: str) -> str:
    """Return a string with the given traceback pretty-printed."""
    pretty_traceback = re.sub(' File ', '\nFile ', traceback)

    return pretty_traceback


# @decorators.map_first_arg
def python_traceback_pretty_print(traceback: str) -> None:
    """Return a string with the given traceback pretty-printed."""
    prettified_string = python_traceback_prettify(traceback)
    print(prettified_string)


# @decorators.map_first_arg
def python_clean(code_text: str) -> str:
    """Clean python code as it is often found in documentation and snippets."""
    code_text = code_text.replace('>>> ', '')
    code_text = code_text.replace('... ', '')
    return code_text


def python_function_blocks(code_text: str, *, ignore_private_functions: bool = False) -> List[str]:
    """Find the code (as a string) for every function in the given code_text."""
    from d8s_lists import has_index
    from d8s_strings import string_chars_at_start_len

    from .ast_data import python_ast_function_defs, python_ast_object_line_numbers

    function_block_strings = []
    code_text_as_lines = code_text.splitlines()
    ast_function_defs = python_ast_function_defs(code_text)
    function_block_line_numbers = [(f.name, python_ast_object_line_numbers(f)) for f in ast_function_defs]

    for function_name, (start, end) in function_block_line_numbers:
        function_block_lines = code_text_as_lines[start - 1 : end]
        function_block_string = '\n'.join(function_block_lines)

        if ignore_private_functions:
            if function_name.startswith('_'):
                continue

        # the code below checks to see if the line after what was determined to be the last line of the function should also be included in the function block (which is the case when the closing parenthesis of a function call in another function is on a newline (see the python_data_tests.py::test_python_function_blocks_edge_cases_1 for an example))
        if has_index(code_text_as_lines, end):
            # find the indentation level of the function definition (the first line of the function)
            function_indentation = string_chars_at_start_len(function_block_lines[0], ' ')
            # TODO: the check below assumes that spaces are used instead of tabs
            next_line_is_indented = (
                code_text_as_lines[end].startswith(' ')
                and string_chars_at_start_len(code_text_as_lines[end], ' ') > function_indentation
            )
            next_line_has_only_parenthesis = code_text_as_lines[end].strip(' ') == ')'
            if next_line_is_indented and next_line_has_only_parenthesis:
                function_block_string += f'\n{code_text_as_lines[end]}'
        function_block_strings.append(function_block_string)
    return function_block_strings


def python_line_count(python_code: str, *, ignore_empty_lines: bool = True) -> int:
    """Return the number of lines in the given function_text."""
    from d8s_lists import truthy_items

    lines = python_code.splitlines()
    if ignore_empty_lines:
        return len(tuple(truthy_items(lines)))
    else:
        return len(lines)


def python_function_lengths(code_text: str) -> List[int]:
    """Find the lengths of each function in the given code_text."""
    function_blocks = python_function_blocks(code_text)
    return [python_line_count(function_block) for function_block in function_blocks]


def python_version() -> str:
    """Return the python version of the current environment."""
    return '{}.{}.{}'.format(sys.version_info.major, sys.version_info.minor, sys.version_info.micro)


def python_is_version_2() -> bool:
    """Return whether or not the python version of the current environment is v2.x."""
    return python_version().startswith('2.')


def python_is_version_3() -> bool:
    """Return whether or not the python version of the current environment is v3.x."""
    return not python_is_version_2()


# TODO: need to standardize the order of arguments between functions like this an the directorySearch function
# @decorators.map_first_arg
def python_files_using_function(function_name: str, search_path: str) -> List[str]:
    """Find where the given function is used in the given search path."""
    from d8s_file_system import directory_read_files_with_path_matching

    files_using_function = []

    # TODO: this function will return the function definition as well as any uses - I would like to split these two things up
    function_pattern = f'{function_name}('
    python_files = directory_read_files_with_path_matching(search_path, '*.py')
    for file_path, file_contents in python_files:
        if function_pattern in file_contents:
            files_using_function.append(file_path)

    return files_using_function


def python_keywords(code_text: Optional[str] = None) -> List[str]:
    """Get a list of the python keywords."""
    import keyword

    python_keywords = keyword.kwlist

    if code_text is None:
        return python_keywords
    else:
        from d8s_strings import string_words

        words_in_code_text = string_words(code_text)
        keywords_used = []

        for palabra in words_in_code_text:
            if palabra in python_keywords:
                keywords_used.append(palabra)

        return keywords_used


# @decorators.map_first_arg
def python_object_properties_enumerate(
    python_object: Any, *, run_methods: bool = True, internal_properties: bool = True
) -> None:
    """Enumerate and print out the properties of the given object."""
    for i in python_object.__dir__():
        if not internal_properties:
            if i.startswith('_'):
                continue

        string_to_eval_as_property = 'python_object.{}'.format(i)
        try:
            if run_methods:
                eval_result = eval(string_to_eval_as_property)
                if callable(eval_result):
                    string_to_eval_as_function = 'python_object.{}()'.format(i)
                    try:
                        print(f'{i}: {eval(string_to_eval_as_function)}')
                    except TypeError:
                        print(f'{i}: {eval(string_to_eval_as_property)}')
                else:
                    print(f'{i}: {eval_result}')
            else:
                print(f'{i}: {eval(string_to_eval_as_property)}')
        except AttributeError:
            print(f'! Unable to get the {i} attribute for the item.')


def python_copy_deep(python_object: Any) -> Any:
    """Return a deep (complete, recursive) copy of the given python object."""
    import copy

    return copy.deepcopy(python_object)


def python_copy_shallow(python_object: Any) -> Any:
    """Return shallow copy of the given python object."""
    import copy

    return copy.copy(python_object)


# @decorators.map_first_arg
def python_file_names(path: str, *, exclude_tests: bool = False) -> List[str]:
    """Find all python files in the given directory."""
    from d8s_file_system import directory_file_names_matching

    files = directory_file_names_matching(path, '*.py')

    if not exclude_tests:
        return files
    else:
        non_test_files = []

        for file in files:
            if '_test' not in file and 'test_' not in file:
                non_test_files.append(file)

        return non_test_files


# @decorators.map_first_arg
def python_fstrings(code_text: str, *, include_braces: bool = False) -> Iterator[str]:
    """Find all of the python formatted string literals in the given text. See https://realpython.com/python-f-strings/ for more details about f-strings."""
    from d8s_grammars import python_formatted_string_literal
    from d8s_lists import flatten

    python_f_strings = flatten(python_formatted_string_literal.searchString(code_text).asList())

    if not include_braces:
        python_f_strings = (f_string.strip('{').strip('}') for f_string in python_f_strings)

    return python_f_strings


# @decorators.map_first_arg
def python_code_details(code_text: str):
    """Get details about the given code_text. This is a wrapper for `dis.code_info`"""
    import dis

    return dis.code_info(code_text)


# @decorators.map_first_arg
def python_disassemble(code_text: str):
    """Disassemble the python code_text. This is a wrapper for `dis.dis`"""
    import dis

    return dis.Bytecode(code_text).dis()


def python_stack_local_data():
    """Get local data in the current python environment."""
    import inspect

    return inspect.currentframe().f_locals


# @decorators.map_first_arg
def python_object_doc_string(python_object: Any) -> Union[str, None]:
    """Get the doc string for the given python object (e.g. module, function, or class)."""
    import inspect

    return inspect.getdoc(python_object)


# @decorators.map_first_arg
def python_object_source_file(python_object: Any) -> str:
    """Get the source file for the given python object (e.g. module, function, or class)."""
    import inspect

    return inspect.getsourcefile(python_object)


# @decorators.map_first_arg
def python_object_module(python_object: Any) -> str:
    """Get the module for the given python object (e.g. function or class)."""
    return python_object.__module__


# @decorators.map_first_arg
def python_object_source_code(python_object: Any) -> str:
    """Get the source code for the given python object (e.g. module, function, or class)."""
    import inspect

    return inspect.getsource(python_object)


# @decorators.map_first_arg
def python_object_signature(python_object: Any) -> str:
    """Get the argument signature for the given python object (e.g. module, function, or class)."""
    import inspect

    return inspect.signature(python_object)


# TODO: improve the type annotations to be lists of types
def python_sort_type_list_by_name(python_type_list: List[type], **kwargs) -> List[type]:
    """."""
    return sorted(python_type_list, key=lambda x: python_type_name(x), **kwargs)


# @decorators.map_first_arg
def python_type_name(python_type: type) -> str:
    """Return the common name of the given type."""
    return python_type.__name__


def python_object_type_to_word(python_object: Any) -> str:
    """Convert the given python type to a string."""
    return python_type_name(type(python_object))



import ast
import sys
from typing import Iterable, List, Optional, Tuple, Union

import more_itertools

from .ast_data_temp_utils import list_delete_empty_items, list_replace

# TODO: all of these functions where code_text is given should also be able to read a file at a given path (?)


def _python_ast_exception_name(node: Union[ast.Raise, ast.ExceptHandler]) -> str:
    """."""
    if hasattr(node, 'exc') and node.exc:  # this handles ast.Raise nodes
        if hasattr(
            node.exc, 'id'
        ):  # this handles ast.Raise nodes where the exception being raised is an ast.Name (e.g. "e" or "ValueError")
            return node.exc.id
        elif hasattr(
            node.exc.func, 'id'
        ):  # this handles ast.Raise nodes where the exception being raised is an ast.Call (e.g. "ValueError('Foo Bar')")
            return node.exc.func.id
        elif hasattr(
            node.exc.func, 'attr'
        ):  # this handles ast.Raise nodes raising a non-built-in error (e.g. "pint.UndefinedUnitError")
            return f'{node.exc.func.value.id}.{node.exc.func.attr}'
    elif hasattr(node, 'type') and node.type:  # this handles ast.ExceptHandler nodes
        if hasattr(
            node.type, 'id'
        ):  # this handles ast.ExceptHandler nodes raising a built-in error (e.g. "RuntimeError")
            return node.type.id
        elif hasattr(
            node.type, 'attr'
        ):  # this handles ast.ExceptHandler nodes raising a non-built-in error (e.g. "pint.UndefinedUnitError")
            return f'{node.type.value.id}.{node.type.attr}'
    elif hasattr(
        node, 'attr'
    ):  # this handles situations where the exception being raised is an ast.Attribute (e.g. "pint.UndefinedUnitError")
        return f'{node.value.id}.{node.attr}'
    elif hasattr(node, 'id'):  # this handles situations where the exception being raised is an ast.Name (e.g. "e")
        return node.id


def python_ast_raise_name(node: ast.Raise) -> Optional[str]:
    """Get the name of the exception raise by the given ast.Raise object."""
    if isinstance(node, ast.Raise):
        return _python_ast_exception_name(node)


def python_ast_exception_handler_exceptions_handled(handler: ast.ExceptHandler) -> Optional[Iterable[str]]:
    """Return all of the exceptions handled by the given exception handler."""
    if isinstance(handler, ast.ExceptHandler):
        handler_has_multiple_exceptions = handler.type and hasattr(handler.type, 'elts')
        if handler_has_multiple_exceptions:
            yield from (_python_ast_exception_name(i) for i in handler.type.elts)
        else:
            exception_name = _python_ast_exception_name(handler)
            if exception_name:
                yield exception_name


def python_ast_exception_handler_exceptions_raised(handler: ast.ExceptHandler) -> Optional[Iterable[str]]:
    """Return the exception raised by the given exception handler."""
    if isinstance(handler, ast.ExceptHandler):
        raise_nodes = list(python_ast_objects_of_type(handler, ast.Raise))
        if any(raise_nodes):
            exceptions_names = list(map(python_ast_raise_name, raise_nodes))
            for name in exceptions_names:
                if name and name == handler.name:
                    exceptions_names = list_replace(
                        exceptions_names, name, python_ast_exception_handler_exceptions_handled(handler)
                    )
                elif name is None:
                    exceptions_names = list_replace(
                        exceptions_names, name, python_ast_exception_handler_exceptions_handled(handler)
                    )
            yield from more_itertools.collapse(exceptions_names, base_type=str)


def python_exceptions_handled(code_text: str) -> Iterable[str]:
    """Return a list of all exceptions handled in the given code."""
    ast_except_handlers = python_ast_objects_of_type(code_text, ast.ExceptHandler)
    yield from more_itertools.collapse(
        list(map(python_ast_exception_handler_exceptions_handled, ast_except_handlers)), base_type=str
    )


def python_exceptions_raised(code_text: str) -> Iterable[str]:
    """Return a list of all exceptions raised in the given code."""
    parsed_code = python_ast_parse(code_text)

    ast_except_handlers = python_ast_objects_of_type(parsed_code, ast.ExceptHandler)
    exceptions = list(map(python_ast_exception_handler_exceptions_raised, ast_except_handlers))

    # remove all of the ast.ExceptHandlers so exceptions are not parsed twice (once from the code above and once in the code below)
    nodes = python_ast_objects_not_of_type(parsed_code, ast.ExceptHandler)
    exceptions.extend(list(map(python_ast_raise_name, (node for node in nodes if isinstance(node, ast.Raise)))))

    yield from more_itertools.collapse(exceptions, base_type=str)


def python_functions_as_import_string(code_text: str, module_name: str) -> str:
    """."""
    import jinja2

    function_names = python_function_names(code_text)
    template = '''from {{ module_name }} import (
{%- for name in function_names %}
    {{ name }},
{%- endfor %}
)'''
    template = jinja2.Template(template)
    result = template.render(module_name=module_name, function_names=function_names)
    return result


def python_ast_object_line_number(ast_object: object) -> Optional[int]:
    """."""
    if hasattr(ast_object, 'lineno'):
        return ast_object.lineno
    else:
        return None


def python_ast_object_line_numbers(ast_object: object) -> Tuple[int, int]:
    """."""
    from d8s_algorithms import depth_first_traverse

    line_numbers = list_delete_empty_items(
        list(
            depth_first_traverse(ast_object, ast.iter_child_nodes, collect_items_function=python_ast_object_line_number)
        )
    )
    return min(line_numbers), max(line_numbers)


def _python_ast_clean(code_text: str) -> str:
    """."""
    import re

    return re.sub('\n', '\\\\n', code_text)


# TODO: have a decorator to parse a first argument that is a string
def python_ast_objects_of_type(code_text_or_ast_object: Union[str, object], ast_type: type) -> Iterable[object]:
    """Return all of the ast objects of the given ast_type in the code_text_or_ast_object."""
    if isinstance(code_text_or_ast_object, str):
        parsed_code = python_ast_parse(code_text_or_ast_object)
    else:
        parsed_code = code_text_or_ast_object
    ast_objects_of_type = (node for node in ast.walk(parsed_code) if isinstance(node, ast_type))
    return ast_objects_of_type


def python_ast_objects_not_of_type(code_text_or_ast_object: Union[str, object], ast_type: type) -> Iterable[object]:
    """Return all of the ast objects which are not of the given ast_type in the code_text_or_ast_object."""
    from d8s_algorithms import depth_first_traverse

    if isinstance(code_text_or_ast_object, str):
        parsed_code = python_ast_parse(code_text_or_ast_object)
    else:
        parsed_code = code_text_or_ast_object

    ast_objects_not_of_type = list_delete_empty_items(
        list(
            depth_first_traverse(
                parsed_code,
                lambda x: ast.iter_child_nodes(x) if not isinstance(x, ast_type) else [],
                collect_items_function=lambda x: x,
            )
        )
    )
    return ast_objects_not_of_type

    # ast_objects_not_of_type = (node for node in ast.walk(parsed_code) if not isinstance(node, ast_type))
    # for node in ast.walk(parsed_code):
    #     if not isinstance(node, ast_type):
    #         yield node
    #     else:
    #         break

    # return ast_objects_not_of_type


def python_ast_parse(code_text: str) -> ast.Module:
    """."""
    # TODO: not sure if this is the right thing to do, but I am cleaning the text before parsing; may want to do this only after I try to parse it without modification the first time
    _python_ast_clean(code_text)
    parsed_code = ast.parse(code_text)
    return parsed_code


def python_ast_function_defs(code_text: str) -> Iterable[ast.FunctionDef]:
    """."""
    function_defs = python_ast_objects_of_type(code_text, ast.FunctionDef)
    return function_defs


def python_function_arguments(function_text: str) -> List[ast.arg]:
    """."""
    parsed_code = python_ast_parse(function_text)
    args = parsed_code.body[0].args.args
    return args


def python_function_argument_names(function_text: str) -> Iterable[str]:
    """."""
    argument_names = (arg.arg for arg in python_function_arguments(function_text))
    return argument_names


def python_function_argument_defaults(function_text: str) -> List[str]:
    """."""
    # TODO: this function does not return defaults for keyword args
    parsed_code = python_ast_parse(function_text)
    return parsed_code.body[0].args.defaults


def python_function_argument_annotations(function_text: str) -> List[str]:
    """."""
    annotations = []
    args = python_function_arguments(function_text)
    for arg in args:
        if arg.annotation:
            annotations.append(arg.annotation.id)
        else:
            annotations.append(None)
    return annotations


# TODO: implement this
# def python_function_keyword_argument_names(function_text: str):
#     """."""
#     pass


def python_function_names(code_text: str, *, ignore_private_functions: bool = False) -> List[str]:
    """."""
    function_objects = python_ast_function_defs(code_text)
    function_names = [f.name for f in function_objects]
    if ignore_private_functions:
        function_names = [name for name in function_names if not name.startswith('_')]
    return function_names


# TODO: write functions to get docstrings for classes and modules
def python_function_docstrings(code_text: str, *, ignore_private_functions: bool = False) -> List[str]:
    """Get docstrings for all of the functions in the given text."""
    function_objects = python_ast_function_defs(code_text)
    docstrings = [
        ast.get_docstring(f) for f in function_objects if not (ignore_private_functions and f.name.startswith('_'))
    ]
    return docstrings


def python_variable_names(code_text: str) -> List[str]:
    """Get all of the variables names in the code_text."""
    # TODO: add a caveat that this function will only find *stored* variables and not those which are referenced or loaded. E.g., given "x = y + 1", this function will return ["x"]; note that "y" is not included
    parsed_code = python_ast_parse(code_text)
    variable_names = [
        node.id for node in ast.walk(parsed_code) if isinstance(node, ast.Name) and (isinstance(node.ctx, ast.Store))
    ]
    return variable_names


def python_constants(code_text: str) -> List[str]:
    """Get all constants in the code_text."""
    # TODO: add a caveat that this function will only find *stored* variables which are uppercased
    variables = python_variable_names(code_text)
    constants = [var for var in variables if var.isupper()]
    return constants



import inspect
import os

import pytest
from d8s_file_system import directory_create, directory_delete, file_name, file_read, file_write

from d8s_python import (
    python_clean,
    python_code_details,
    python_copy_deep,
    python_copy_shallow,
    python_disassemble,
    python_file_names,
    python_files_using_function,
    python_fstrings,
    python_function_blocks,
    python_function_lengths,
    python_functions_signatures,
    python_is_version_2,
    python_is_version_3,
    python_keywords,
    python_line_count,
    python_make_pythonic,
    python_namespace_has_argument,
    python_object_doc_string,
    python_object_module,
    python_object_properties_enumerate,
    python_object_signature,
    python_object_source_code,
    python_object_source_file,
    python_object_type_to_word,
    python_sort_type_list_by_name,
    python_stack_local_data,
    python_todos,
    python_traceback_prettify,
    python_traceback_pretty_print,
    python_type_name,
    python_version,
)

PYTHON_FILE_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '../d8s_python/python_data.py'))
PYTHON_FILE_TEXT = file_read(PYTHON_FILE_PATH)
SIMPLE_FUNCTION = '''def test(a) -> List:
    print(a)

    return []'''
DEMOCRITUS_ROOT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))


TEST_DIRECTORY_PATH = './test_files'
TEST_FILE_CONTENTS = 'def a():\n    return 1'
TEST_FILE_NAME_1 = 'a.py'
TEST_FILE_NAME_2 = 'a_test.py'
EXISTING_FILE_PATH_1 = os.path.join(TEST_DIRECTORY_PATH, TEST_FILE_NAME_1)
EXISTING_FILE_PATH_2 = os.path.join(TEST_DIRECTORY_PATH, TEST_FILE_NAME_2)


@pytest.fixture(autouse=True)
def clear_testing_directory():
    """This function is run after every test."""
    directory_delete(TEST_DIRECTORY_PATH)
    directory_create(TEST_DIRECTORY_PATH)
    file_write(EXISTING_FILE_PATH_1, TEST_FILE_CONTENTS)
    file_write(EXISTING_FILE_PATH_2, TEST_FILE_CONTENTS)


def setup_module():
    """This function is run before all of the tests in this file are run."""
    directory_create(TEST_DIRECTORY_PATH)


def teardown_module():
    """This function is run after all of the tests in this file are run."""
    directory_delete(TEST_DIRECTORY_PATH)


def test_python_object_signature_1():
    result = python_object_signature(python_object_signature)
    assert isinstance(result, inspect.Signature)


def test_python_object_source_code_1():
    result = python_object_source_code(python_object_signature)
    print(f'result: {result}')
    assert (
        result
        == 'def python_object_signature(python_object: Any) -> str:\n    """Get the argument signature for the given python object (e.g. module, function, or class)."""\n    import inspect\n\n    return inspect.signature(python_object)\n'
    )


def test_python_function_lengths_1():
    result = python_function_lengths(SIMPLE_FUNCTION)
    assert result == [3]


def test_python_line_count_1():
    result = python_line_count(SIMPLE_FUNCTION)
    assert result == 3

    result = python_line_count(SIMPLE_FUNCTION, ignore_empty_lines=False)
    assert result == 4


def test_python_todos_1():
    s = '''# TODO: hi!'''
    result = python_todos(s)
    assert result == ['TODO: hi!']

    s = '''# todo: hi!'''
    result = python_todos(s)
    assert result == []

    s = '''# todo: hi!'''
    result = python_todos(s, 'todo:.*')
    assert result == ['todo: hi!']


def test_python_stack_local_data_docs_1():
    result = python_stack_local_data()
    assert isinstance(result, dict)
    assert 'inspect' in result


def test_python_code_details_docs_1():
    result = python_code_details(SIMPLE_FUNCTION)
    assert result.startswith(
        '''Name:              <module>
Filename:          <disassembly>
'''
    )


def test_python_disassemble_docs_1():
    result = python_disassemble(SIMPLE_FUNCTION)
    print(f'result {result} ')
    assert result.strip().startswith(
        '''1           0 LOAD_NAME                0 (List)
              2 LOAD_CONST               0 (('return',))
              4 BUILD_CONST_KEY_MAP      1
              6 LOAD_CONST               1'''
    )


def test_python_copy_deep_docs_1():
    # make a list, make a copy of the list, modify the original list and the copy, and make sure that the lists are modified independently of one another
    l = [2, 3, 5, 7]
    result = python_copy_deep(l)
    l.pop(0)
    result.append(11)
    assert l == [3, 5, 7]
    assert result == [2, 3, 5, 7, 11]


def test_python_object_properties_enumerate_docs_1(capsys):
    s = 'foo'
    python_object_properties_enumerate(s)
    captured = capsys.readouterr()
    assert captured.out
    assert 'upper: FOO' in captured.out
    assert "__getattribute__: <method-wrapper '__getattribute__' of str object" in captured.out

    s = 'foo'
    python_object_properties_enumerate(s, run_methods=False)
    captured = capsys.readouterr()
    assert captured.out
    assert 'upper: FOO' not in captured.out
    assert "__getattribute__: <method-wrapper '__getattribute__' of str object" in captured.out

    s = 'foo'
    python_object_properties_enumerate(s, internal_properties=False)
    captured = capsys.readouterr()
    assert captured.out
    assert 'upper: FOO' in captured.out
    assert "__getattribute__: <method-wrapper '__getattribute__' of str object" not in captured.out

    s = 'foo'
    python_object_properties_enumerate(s, run_methods=False, internal_properties=False)
    captured = capsys.readouterr()
    assert captured.out
    assert 'upper: FOO' not in captured.out
    assert "__getattribute__: <method-wrapper '__getattribute__' of str object" not in captured.out


def test_python_version_docs_1():
    result = python_version()
    assert isinstance(result, str)
    assert result.startswith('3.')


def test_python_clean_docs_1():
    s = '''>>> import argparse
>>> args = argparse.Namespace()
>>> args.foo = 1'''
    result = python_clean(s)
    assert (
        result
        == '''import argparse
args = argparse.Namespace()
args.foo = 1'''
    )


def test_python_namespace_has_argument_docs_1():
    import argparse

    namespace = argparse.Namespace()
    namespace.foo = 1
    assert python_namespace_has_argument(namespace, 'foo')
    assert not python_namespace_has_argument(namespace, 'bar')


def test_python_object_doc_string_docs_1():
    def a():
        """Foo bar."""
        return 1

    result = python_object_doc_string(a)
    assert result == 'Foo bar.'

    def a():
        return 1

    result = python_object_doc_string(a)
    assert result == None


def test_python_functions_signatures_docs_1():
    s = '''def string_remove_numbers(input_string: str, replacement: str = ' ') -> str:
    """Remove all numbers from the input_strings."""
    new_string_without_numbers = replace('\d+', replacement, input_string)
    return new_string_without_numbers'''
    results = python_functions_signatures(s)
    print(f'results {results} ')
    assert results == ["(input_string: str, replacement: str = ' ') -> str"]

    s = '''def python_functions_signatures(
    code_text: str, ignore_private_functions: bool = False, keep_function_name: bool = False
) -> ListOfStrs:
    print('hi')'''
    results = python_functions_signatures(s)
    assert results == [
        '(\n    code_text: str, ignore_private_functions: bool = False, keep_function_name: bool = False\n) -> ListOfStrs'
    ]

    s = '''@decorators.map_first_arg
def indefinite_article(word):
    """Return the word(s) with the appropriate indefinite article."""
    inflect_engine = _inflect_engine()
    return inflect_engine.a(word).split(' ')[0]


@decorators.map_first_arg
def plural(word, count=None):
    """Make the word(s) plural."""
    inflect_engine = _inflect_engine()
    return inflect_engine.plural(word, count=count)'''
    results = python_functions_signatures(s)
    print(f'results {results} ')
    assert results == ['(word)', '(word, count=None)']

    results = python_functions_signatures(SIMPLE_FUNCTION)
    print(f'results {results} ')
    assert results == ['(a) -> List']

    results = python_functions_signatures(SIMPLE_FUNCTION, keep_function_name=True)
    print(f'results {results} ')
    assert results == ['test(a) -> List']


def test_python_traceback_prettify_docs_1():
    traceback = '''File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/exception.py", line 41, in inner response = get_response(request) File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/base.py", line 187, in _get_response response = self.process_exception_by_middleware(e, request) File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/base.py", line 185, in _get_response response = wrapped_callback(request, *callback_args, **callback_kwargs)'''
    pretty_traceback = python_traceback_prettify(traceback)
    print('pretty_traceback {}'.format(pretty_traceback))
    assert (
        pretty_traceback
        == '''File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/exception.py", line 41, in inner response = get_response(request)
File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/base.py", line 187, in _get_response response = self.process_exception_by_middleware(e, request)
File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/base.py", line 185, in _get_response response = wrapped_callback(request, *callback_args, **callback_kwargs)'''
    )


def test_python_traceback_pretty_print_docs_1(capsys):
    traceback = '''File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/exception.py", line 41, in inner response = get_response(request) File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/base.py", line 187, in _get_response response = self.process_exception_by_middleware(e, request) File "/app/.heroku/python/lib/python3.6/site-packages/django/core/handlers/base.py", line 185, in _get_response response = wrapped_callback(request, *callback_args, **callback_kwargs)'''
    pretty_traceback = python_traceback_prettify(traceback)
    python_traceback_pretty_print(traceback)
    captured = capsys.readouterr()
    assert captured.out.strip() == pretty_traceback.strip()


def test_python_fstrings_docs_1():
    python_f_strings = tuple(python_fstrings('Hello, {name}. You are {age}.'))
    assert python_f_strings == ('name', 'age')

    python_f_strings = tuple(python_fstrings('Hello, {name}. You are {age}.', include_braces=True))
    assert python_f_strings == ('{name}', '{age}')

    python_f_strings = tuple(python_fstrings('Hello, {new_comedian!r}'))
    assert python_f_strings == ('new_comedian!r',)

    python_f_strings = tuple(python_fstrings('Hello, {}'.format('Bob')))
    assert python_f_strings == ()


def test_python_object_type_to_word_docs_1():
    assert python_object_type_to_word(1) == 'int'
    assert python_object_type_to_word('foo') == 'str'
    assert python_object_type_to_word([]) == 'list'
    assert python_object_type_to_word({}) == 'dict'
    assert python_object_type_to_word({1, 2, 3}) == 'set'
    assert python_object_type_to_word(('foo', 'bar')) == 'tuple'
    assert python_object_type_to_word(True) == 'bool'
    assert python_object_type_to_word(2.0) == 'float'

    class Animal:
        pass

    a = Animal()
    assert python_object_type_to_word(a) == 'Animal'


def test_python_make_pythonic_docs_1():
    assert python_make_pythonic('fooBar') == 'foo_bar'
    assert python_make_pythonic('foo-bar') == 'foo_bar'
    assert python_make_pythonic('foo bar') == 'foo_bar'
    assert python_make_pythonic('Foo Bar') == 'foo_bar'
    # make sure that consecutive capital letters are not split up
    assert python_make_pythonic('ET Phone Home') == 'et_phone_home'
    assert python_make_pythonic('IP address data') == 'ip_address_data'


def test_python_object_source_file_docs_1():
    result = python_object_source_file(file_read)
    assert result.endswith('d8s_file_system/files.py')


def test_python_object_module_docs_1():
    result = python_object_module(file_read)
    assert result == 'd8s_file_system.files'


def test_python_sort_type_list_by_name_docs_1():
    result = python_sort_type_list_by_name([str, int, tuple, float])
    assert result == [float, int, str, tuple]

    result = python_sort_type_list_by_name([str, int, tuple, float], reverse=True)
    assert result == [tuple, str, int, float]


def test_python_file_names_docs_1():
    # TODO: create a directory of data locally that we test against
    assert python_file_names(TEST_DIRECTORY_PATH) == ['a.py', 'a_test.py']
    assert python_file_names(TEST_DIRECTORY_PATH, exclude_tests=True) == ['a.py']


def test_python_files_using_function_docs_1():
    results = python_files_using_function('python_files_using_function', search_path=DEMOCRITUS_ROOT_PATH)
    assert len(results) == 2
    assert file_name(results[0]) == 'test_python_data.py' or file_name(results[1]) == 'test_python_data.py'


def test_python_is_version_2_docs_1():
    assert python_is_version_2() == False


def test_python_is_version_3_docs_1():
    assert python_is_version_3() == True


def test_python_keywords_docs_1():
    assert python_keywords() == [
        'False',
        'None',
        'True',
        'and',
        'as',
        'assert',
        'async',
        'await',
        'break',
        'class',
        'continue',
        'def',
        'del',
        'elif',
        'else',
        'except',
        'finally',
        'for',
        'from',
        'global',
        'if',
        'import',
        'in',
        'is',
        'lambda',
        'nonlocal',
        'not',
        'or',
        'pass',
        'raise',
        'return',
        'try',
        'while',
        'with',
        'yield',
    ]
    assert python_keywords(code_text=SIMPLE_FUNCTION) == ['def', 'return']


def test_python_object_properties_enumerate_docs_1():
    assert python_object_properties_enumerate('foo') == None


def test_python_function_blocks_1():
    result = python_function_blocks(SIMPLE_FUNCTION)
    assert result == ['def test(a) -> List:\n    print(a)\n\n    return []']

    s = '''def a(
    n: str, o: Callable
):
    print('foo')'''
    assert python_function_blocks(s) == [s]

    s = '''def a(
    n: str, o: Callable
):
    print("""in the beginning
was the word
and the word...""")
    x = 10
    if x == 1:
        x = 1
    elif x == 2:
        x = 2
    else:
      print('bar')


a = 1
b = 2
c = 3
myList = range(10)

def someMethod(x):
    something = x * 2
    def subfunction(something):
        print('here')
    return something

f = someMethod(b)

print(f)'''
    assert python_function_blocks(s) == [
        'def a(\n    n: str, o: Callable\n):\n    print("""in the beginning\nwas the word\nand the word...""")\n    x = 10\n    if x == 1:\n        x = 1\n    elif x == 2:\n        x = 2\n    else:\n      print(\'bar\')',
        "def someMethod(x):\n    something = x * 2\n    def subfunction(something):\n        print('here')\n    return something",
        "    def subfunction(something):\n        print('here')",
    ]

    s = '''@do.something
def _a(
    n: str, o: Callable
):
    print('foo')'''
    assert python_function_blocks(s) == [s]
    assert python_function_blocks(s, ignore_private_functions=True) == []


def test_python_function_blocks_edge_cases_1():
    # when this tests was written, the python_function_blocks function wasn't finding anything after "b(\n        1"
    s = '''def a():
    def b(n):
        return (
            n * 2
        )
        
    return b(
        1
    )'''
    assert python_function_blocks(s) == [
        'def a():\n    def b(n):\n        return (\n            n * 2\n        )\n        \n    return b(\n        1\n    )',
        '    def b(n):\n        return (\n            n * 2\n        )',
    ]

    s = '''def a():
    return some_collection.get_objects(locator=l5) \
                      .get_distinct(case_insensitive=True) \
                      .filter(predicate=query(q5)) \
                      .values()'''
    assert python_function_blocks(s) == [
        'def a():\n    return some_collection.get_objects(locator=l5)                       .get_distinct(case_insensitive=True)                       .filter(predicate=query(q5))                       .values()'
    ]

    s = '''def a():
    return "foo" +\
    "bar"'''
    print(f'python_function_blocks(s): {python_function_blocks(s)}')
    assert python_function_blocks(s) == ['def a():\n    return "foo" +    "bar"']



import ast
import os

from d8s_file_system import file_read

from d8s_python import (
    python_ast_exception_handler_exceptions_raised,
    python_ast_function_defs,
    python_ast_object_line_numbers,
    python_ast_objects_not_of_type,
    python_ast_objects_of_type,
    python_ast_parse,
    python_constants,
    python_exceptions_handled,
    python_exceptions_raised,
    python_function_argument_annotations,
    python_function_argument_defaults,
    python_function_argument_names,
    python_function_arguments,
    python_function_docstrings,
    python_function_names,
    python_functions_as_import_string,
    python_variable_names,
)
from d8s_python.ast_data import _python_ast_clean

TEST_CODE = '''

async def foo():
    """."""
    print('Here!')

def python_function_names(code_text: str, *, ignore_private_functions: bool = False) -> List[str]:
    """."""
    function_objects = python_ast_function_defs(code_text)
    function_names = [f.name for f in function_objects]
    if ignore_private_functions:
        function_names = [name for name in function_names if not name.startswith('_')]
    return function_names


# TODO: write functions to get docstrings for classes and modules
def python_function_docstrings(code_text: str, *, ignore_private_functions: bool = False) -> List[str]:
    """Get docstrings for all of the functions in the given text."""
    function_objects = python_ast_function_defs(code_text)
    docstrings = [
        ast.get_docstring(f) for f in function_objects if not (ignore_private_functions and f.name.startswith('_'))
    ]
    return docstrings


def python_variable_names(code_text: str) -> List[str]:
    """Get all of the variables names in the code_text."""
    # TODO: add a caveat that this function will only find *stored* variables and not those which are referenced or loaded. E.g., given "x = y + 1", this function will return ["x"]; note that "y" is not included
    parsed_code = python_ast_parse(code_text)
    variable_names = [
        node.id for node in ast.walk(parsed_code) if isinstance(node, ast.Name) and (isinstance(node.ctx, ast.Store))
    ]
    return variable_names


def python_constants(code_text: str) -> List[str]:
    """Get all constants in the code_text."""
    # TODO: add a caveat that this function will only find *stored* variables which are uppercased
    variables = python_variable_names(code_text)
    constants = [var for var in variables if var.isupper()]
    return constants

'''

TEST_CODE_1 = '''a = 1
b = 2
c = 3
myList = range(10)

def someMethod(x):
    something = x * 2
    return something

f = someMethod(b)

print(f)'''

TEST_CODE_WITH_EXCEPTION = '''try:
    return a  / b
except ZeroDivisionError:
    raise'''

TEST_CODE_WITH_PRIVATE_FUNCTION = '''def _test(a: str):
    """Docstring."""
    return a'''

TEST_FUNCTION_WITH_DEFAULT = '''def test(a: str = ''):
    """Docstring."""
    return a'''

TEST_EXCEPTION_DATA = [
    {  # test a simple raise
        'code': '''raise ValueError('I cannot divide by zero')''',
        'handled': [],
        'raised': ['ValueError'],
    },
    {  # test a simple raise (in a context)
        'code': '''if d == 0:
    raise ValueError('I cannot divide by zero')
else:
    return n / d''',
        'handled': [],
        'raised': ['ValueError'],
    },
    {  # test code where an error that is part of another module is raised
        'code': '''raise pint.UndefinedUnitError('"Foo" is not a valid unit')''',
        'handled': [],
        'raised': ['pint.UndefinedUnitError'],
    },
    {  # ('one', ('afore_mentioned', ('custom', 'explicit')))
        'code': '''try:\n\tpass\nexcept pint.UndefinedUnitError:\n\traise pint.UndefinedUnitError('"Foo" is not a valid unit')''',
        'handled': ['pint.UndefinedUnitError'],
        'raised': ['pint.UndefinedUnitError'],
    },
    {  # ('one', ('afore_mentioned', ('custom', 'named_before_except')))
        'code': '''e = pint.UndefinedUnitError('"Foo" is not a valid unit')\ntry:\n\tpass\nexcept e:\n\traise e''',
        'handled': ['e'],
        'raised': ['e'],
    },
    {  # ('one', ('afore_mentioned', ('custom', 'named_in_except')))
        'code': '''try:\n\tpass\nexcept pint.UndefinedUnitError as e:\n\traise e''',
        'handled': ['pint.UndefinedUnitError'],
        'raised': ['pint.UndefinedUnitError'],
    },
    {  # ('one', ('afore_mentioned', ('built_in', 'explicit')))
        'code': '''try:\n\tpass\nexcept RuntimeError:\n\traise RuntimeError('"Foo" is not a valid unit')''',
        'handled': ['RuntimeError'],
        'raised': ['RuntimeError'],
    },
    {  # ('one', ('afore_mentioned', ('built_in', 'named_before_except')))
        'code': '''e = ValueError('"Foo" is not a valid unit')\ntry:\n\tpass\nexcept e:\n\traise e''',
        'handled': ['e'],
        'raised': ['e'],
    },
    {  # ('one', ('afore_mentioned', ('built_in', 'named_in_except')))
        'code': '''try:\n\tpass\nexcept RuntimeError as e:\n\traise e''',
        'handled': ['RuntimeError'],
        'raised': ['RuntimeError'],
    },
    {  # ('one', ('different', ('custom', 'explicit')))
        'code': '''try:\n\tpass\nexcept RuntimeError:\n\traise pint.UndefinedUnitError('"Foo" is not a valid unit')''',
        'handled': ['RuntimeError'],
        'raised': ['pint.UndefinedUnitError'],
    },
    {  # ('one', ('different', ('custom', 'named_before_except')))
        'code': '''e = pint.UndefinedUnitError('"Foo" is not a valid unit')\ntry:\n\tpass\nexcept ValueError:\n\traise e''',
        'handled': ['ValueError'],
        'raised': ['e'],
    },
    {  # ('one', ('different', ('built_in', 'explicit')))
        'code': '''try:\n\tpass\nexcept ValueError:\n\traise RuntimeError''',
        'handled': ['ValueError'],
        'raised': ['RuntimeError'],
    },
    {  # ('one', ('different', ('built_in', 'named_before_except')))
        'code': '''e = ValueError\ntry:\n\tpass\nexcept RuntimeError:\n\traise e''',
        'handled': ['RuntimeError'],
        'raised': ['e'],
    },
    {  # ('one', '') - with built-in exception
        'code': '''try:\n\tpass\nexcept RuntimeError:\n\traise''',
        'handled': ['RuntimeError'],
        'raised': ['RuntimeError'],
    },
    {  # ('one', '') - with custom exception
        'code': '''try:\n\tpass\nexcept pint.UndefinedUnitError:\n\traise''',
        'handled': ['pint.UndefinedUnitError'],
        'raised': ['pint.UndefinedUnitError'],
    },
    {  # ('many', ('afore_mentioned', ('custom', 'explicit')))
        'code': '''try:\n\tpass\nexcept (pint.UndefinedUnitError, pint.FooBarError):\n\traise pint.UndefinedUnitError('"Foo" is not a valid unit')''',
        'handled': ['pint.UndefinedUnitError', 'pint.FooBarError'],
        'raised': ['pint.UndefinedUnitError'],
    },
    {  # ('many', ('afore_mentioned', ('custom', 'named_before_except')))
        'code': '''e = pint.UndefinedUnitError('"Foo" is not a valid unit')\ntry:\n\tpass\nexcept (e, pint.FooBarError):\n\traise e''',
        'handled': ['e', 'pint.FooBarError'],
        'raised': ['e'],
    },
    {  # ('many', ('afore_mentioned', ('custom', 'named_in_except')))
        'code': '''try:\n\tpass\nexcept (pint.UndefinedUnitError, pint.FooBarError) as e:\n\traise e''',
        'handled': ['pint.UndefinedUnitError', 'pint.FooBarError'],
        'raised': ['pint.UndefinedUnitError', 'pint.FooBarError'],
    },
    {  # ('many', ('afore_mentioned', ('built_in', 'explicit')))
        'code': '''try:\n\tpass\nexcept (RuntimeError, RuntimeWarning):\n\traise RuntimeError('"Foo" is not a valid unit')''',
        'handled': ['RuntimeError', 'RuntimeWarning'],
        'raised': ['RuntimeError'],
    },
    {  # ('many', ('afore_mentioned', ('built_in', 'named_before_except')))
        'code': '''e = ValueError('"Foo" is not a valid unit')\ntry:\n\tpass\nexcept (e, RuntimeError):\n\traise e''',
        'handled': ['e', 'RuntimeError'],
        'raised': ['e'],
    },
    {  # ('many', ('afore_mentioned', ('built_in', 'named_in_except')))
        'code': '''try:\n\tpass\nexcept (RuntimeError, RuntimeWarning) as e:\n\traise e''',
        'handled': ['RuntimeError', 'RuntimeWarning'],
        'raised': ['RuntimeError', 'RuntimeWarning'],
    },
    {  # ('many', ('different', ('custom', 'explicit')))
        'code': '''try:\n\tpass\nexcept (pint.AError, pint.BError):\n\traise pint.UndefinedUnitError('"Foo" is not a valid unit')''',
        'handled': ['pint.AError', 'pint.BError'],
        'raised': ['pint.UndefinedUnitError'],
    },
    {  # ('many', ('different', ('custom', 'named_before_except')))
        'code': '''e = pint.UndefinedUnitError('"Foo" is not a valid unit')\ntry:\n\tpass\nexcept (ValueError, RuntimeError):\n\traise e''',
        'handled': ['ValueError', 'RuntimeError'],
        'raised': ['e'],
    },
    {  # ('many', ('different', ('built_in', 'explicit')))
        'code': '''try:\n\tpass\nexcept (ValueError, AssertionError):\n\traise RuntimeError''',
        'handled': ['ValueError', 'AssertionError'],
        'raised': ['RuntimeError'],
    },
    {  # ('many', ('different', ('built_in', 'named_before_except')))
        'code': '''e = ValueError\ntry:\n\tpass\nexcept (RuntimeError, RuntimeWarning):\n\traise e''',
        'handled': ['RuntimeError', 'RuntimeWarning'],
        'raised': ['e'],
    },
    {  # ('many', '') - with built-in exception
        'code': '''try:\n\tpass\nexcept (RuntimeError, RuntimeWarning):\n\traise''',
        'handled': ['RuntimeError', 'RuntimeWarning'],
        'raised': ['RuntimeError', 'RuntimeWarning'],
    },
    {  # ('many', '') - with custom exception
        'code': '''try:\n\tpass\nexcept (pint.UndefinedUnitError, pint.AError):\n\traise''',
        'handled': ['pint.UndefinedUnitError', 'pint.AError'],
        'raised': ['pint.UndefinedUnitError', 'pint.AError'],
    },
]


def test_python_ast_exception_handler_exceptions_raised_1():
    except_handler = tuple(python_ast_objects_of_type(TEST_CODE_WITH_EXCEPTION, ast.ExceptHandler))[0]
    result = python_ast_exception_handler_exceptions_raised(except_handler)
    assert tuple(result) == ('ZeroDivisionError',)


def test_python_ast_objects_not_of_type_docs_1():
    # make sure that any ast.FunctionDef objects *and* their children are not returned
    result = python_ast_objects_not_of_type(TEST_CODE_1, ast.FunctionDef)
    assert len(result) == 35
    assert isinstance(result, list)


def test_python_ast_objects_of_type_1():
    result = list(python_ast_objects_of_type(TEST_CODE_1, ast.FunctionDef))
    assert len(result) == 1
    assert isinstance(result[0], ast.FunctionDef)


def test_python_exceptions_handled_docs_1():
    for test in TEST_EXCEPTION_DATA:
        try:
            assert list(python_exceptions_handled(test['code'])) == test['handled']
        except AssertionError as e:
            failure = (test, e)
            print(failure)
            raise e


def test_python_ast_object_line_numbers_docs_1():
    result = python_ast_object_line_numbers(python_ast_parse(TEST_CODE_WITH_PRIVATE_FUNCTION))
    assert result == (1, 3)

    result = python_ast_object_line_numbers(python_ast_parse(TEST_FUNCTION_WITH_DEFAULT))
    assert result == (1, 3)

    s = '''def democritus_directory_documentation_create(directory_path: str, *, track_changes: bool = True):
    """Create documentation files for all non-test python files in the given directory_path."""
    _directory_documentation_action(
        directory_path, democritus_functions_file_documentation_create, track_changes=track_changes
    )'''
    result = python_ast_object_line_numbers(python_ast_parse(s))
    assert result == (1, 4)


def test_python_exceptions_raised_docs_1():
    for test in TEST_EXCEPTION_DATA:
        try:
            assert list(python_exceptions_raised(test['code'])) == test['raised']
        except AssertionError as e:
            failure = (test, e)
            print(failure)
            raise e


def test_python_functions_as_import_string_1():
    print(python_functions_as_import_string(TEST_CODE, 'ast_data'))
    assert (
        python_functions_as_import_string(TEST_CODE, 'ast_data')
        == 'from ast_data import (\n    python_function_names,\n    python_function_docstrings,\n    python_variable_names,\n    python_constants,\n)'
    )


def test__python_ast_clean_1():
    assert _python_ast_clean('print("foo\nbar")') == 'print("foo\\nbar")'


def test_python_ast_parse_1():
    result = python_ast_parse(TEST_CODE_1)
    assert isinstance(result, ast.Module)


def test_python_ast_function_defs_1():
    result = tuple(python_ast_function_defs(TEST_FUNCTION_WITH_DEFAULT))
    assert len(result) == 1
    assert isinstance(result[0], ast.FunctionDef)

    result = tuple(python_ast_function_defs(TEST_CODE_WITH_PRIVATE_FUNCTION))
    assert len(result) == 1
    assert isinstance(result[0], ast.FunctionDef)


def test_python_function_arguments_1():
    result = python_function_arguments(TEST_FUNCTION_WITH_DEFAULT)
    assert len(result) == 1
    assert isinstance(result[0], ast.arg)


def test_python_function_argument_names_1():
    assert tuple(python_function_argument_names(TEST_FUNCTION_WITH_DEFAULT)) == ('a',)


def test_python_function_argument_defaults_1():
    result = python_function_argument_defaults(TEST_FUNCTION_WITH_DEFAULT)
    assert len(result) == 1
    assert isinstance(result[0], ast.Str)

    result = python_function_argument_defaults(TEST_CODE_WITH_PRIVATE_FUNCTION)
    assert len(result) == 0


def test_python_function_argument_annotations_1():
    assert (
        list(
            python_function_argument_annotations(
                '''def _test(a: str):
    """Docstring."""
    return a'''
            )
        )
        == ['str']
    )
    assert (
        list(
            python_function_argument_annotations(
                '''def _test(a):
    """Docstring."""
    return a'''
            )
        )
        == [None]
    )


def test_python_function_names_1():
    assert python_function_names(TEST_CODE) == [
        'python_function_names',
        'python_function_docstrings',
        'python_variable_names',
        'python_constants',
    ]
    assert python_function_names(TEST_CODE_WITH_PRIVATE_FUNCTION) == ['_test']
    assert python_function_names(TEST_CODE_WITH_PRIVATE_FUNCTION, ignore_private_functions=True) == []


def test_python_function_docstrings_1():
    code_text = '''def _test(a: str):
    """Docstring."""
    return a'''
    assert python_function_docstrings(code_text) == ['Docstring.']
    assert python_function_docstrings(code_text, ignore_private_functions=True) == []


def test_python_variable_names_1():
    assert python_variable_names('x = 7') == ['x']
    assert python_variable_names('x = y + 7') == ['x']
    assert python_variable_names('PI = 3.14') == ['PI']
    assert python_variable_names('1 + 0') == []
    assert python_variable_names(TEST_CODE_1) == ['a', 'b', 'c', 'myList', 'f', 'something']


def test_python_constants_1():
    assert python_constants('x = 7') == []
    assert python_constants('PI = 3.14') == ['PI']
    assert python_constants('1 + 0') == []



from typing import Any, Iterable, Tuple


def list_item_indexes(list_arg: list, item: Any) -> Tuple[int, ...]:
    """Find the given item in the iterable. Return -1 if the item is not found."""
    indexes = [index for index, value in enumerate(list_arg) if value == item]
    return indexes


def list_delete_item(list_arg: list, item_to_delete: Any) -> list:
    """Remove all instances of the given item_to_delete from the list_arg."""
    from itertools import filterfalse

    result = list(filterfalse(lambda x: x == item_to_delete, list_arg))
    return result


def list_replace(list_arg: list, old_value, new_value, *, replace_in_place: bool = True) -> list:
    """Replace all instances of the old_value with the new_value in the given list_arg."""
    old_value_indexes = list_item_indexes(list_arg, old_value)
    new_list = list_delete_item(list_arg, old_value)

    for index in old_value_indexes:
        if replace_in_place:
            new_list.insert(index, new_value)
        else:
            new_list.append(new_value)

    return new_list


def list_delete_empty_items(list_arg: list) -> list:
    """Delete items from the list_arg is the item is an empty strings, empty list, zero, False or None."""
    empty_values = ('', [], 0, False, None)
    # TODO: not sure if this is the right way to implement this
    return [i for i in list_arg if i not in empty_values]
